package colorspace

import "math"

type TransferFunction func(e float64) float64

const DCINormalizationFactor = 48 / 52.37

const (
	Gamma22      = 2.2
	Gamma24      = 2.4
	GammaDCIXYZ  = 2.6
	GammaSRGB    = Gamma22
	GammaRec709  = 1 / 0.45
	GammaRec2020 = GammaRec709
)

// https://www.itu.int/dms_pubrec/itu-r/rec/bt/R-REC-BT.2020-2-201510-I!!PDF-E.pdf
const alpha = 1.09929682680944 // 10 * Math.pow(beta, 0.55)
const beta = 0.018053968510807

var (
	TransferFromDCIXYZ TransferFunction = func(e float64) float64 {
		return e / DCINormalizationFactor
	}
	CompandingSRGB TransferFunction = func(e float64) float64 {
		if e <= 0.0031308 {
			return 12.92 * e
		} else {
			return 1.055*math.Pow(e, 1/GammaSRGB) - 0.055
		}
	}

	// CompandingRec709 ITU-R BT.709
	CompandingRec709 TransferFunction = func(e float64) float64 {
		if e < beta {
			return 4.5 * e
		} else {
			return alpha*math.Pow(e, 1/GammaRec709) - (alpha - 1)
		}
	}
	PureRec709 TransferFunction = func(e float64) float64 {
		return math.Pow(e, 1/GammaRec709)
	}
	PureRec709_22 TransferFunction = func(e float64) float64 {
		return math.Pow(e, 1/Gamma22)
	}
	PureRec709_24 TransferFunction = func(e float64) float64 {
		return math.Pow(e, 1/Gamma24)
	}
	CompandingRec2020 = CompandingRec709
	PureRec2020       = PureRec709
	PureRec2020_22    = PureRec709_22
	PureRec2020_24    = PureRec709_24
)

type LUT []float64

func NewGammaLUT(gamma float64, bits int) (lut LUT) {
	size := (1 << bits) - 1
	lut = make(LUT, size+1)
	for i := 0; i <= size; i++ {
		lut[i] = math.Pow(float64(i)/float64(size), gamma)
	}
	return lut
}
