package libav

/*
#cgo pkg-config: libavformat libavcodec libavutil

#include <libavformat/avformat.h>
*/
import "C"
import (
	"errors"
	"io"
	"slices"
	"unsafe"
)

type PacketData struct {
	Number int
	Data   []byte
}

func OpenXYZ12(inputFile string, initFunc func(framerateNum, framerateDen, sarNum, sarDen, width, height int) error, packetFunc func(packet PacketData) error) error {
	var fmtCtx *C.AVFormatContext

	//Open file and decoder

	inputFileC := C.CString(inputFile)
	defer C.free(unsafe.Pointer(inputFileC))

	ret := C.avformat_open_input(&fmtCtx, inputFileC, nil, nil)
	if ret < 0 {
		panic("cannot open input file")
	}
	defer C.avformat_close_input(&fmtCtx)

	ret = C.avformat_find_stream_info(fmtCtx, nil)
	if ret < 0 {
		panic("cannot find stream information")
	}

	//get video
	ret = C.av_find_best_stream(fmtCtx, C.AVMEDIA_TYPE_VIDEO, -1, -1, nil, 0)
	if ret < 0 {
		panic("cannot find video stream")
	}

	videoStreamIndex := ret

	inputStream := unsafe.Slice(fmtCtx.streams, fmtCtx.nb_streams)[videoStreamIndex]

	codecPar := inputStream.codecpar

	if codecPar.codec_id != C.AV_CODEC_ID_JPEG2000 {
		panic("video codec not JPEG2000")
	}

	if codecPar.format != C.AV_PIX_FMT_XYZ12LE {
		panic("video format not xyz12le")
	}

	codecPar.color_range = C.AVCOL_RANGE_JPEG
	codecPar.color_primaries = C.AVCOL_PRI_SMPTE431
	//codecPar.color_trc = C.AVCOL_PRI_SMPTE431
	//codecPar.color_space = C.AVCOL_PRI_SMPTE431

	var packet *C.AVPacket
	packet = C.av_packet_alloc()
	if packet == nil {
		panic("err allocating")
	}
	defer C.av_packet_free(&packet)

	err := initFunc(int(inputStream.codec.framerate.num), int(inputStream.codec.framerate.den), int(inputStream.sample_aspect_ratio.num), int(inputStream.sample_aspect_ratio.den), int(codecPar.width), int(codecPar.height))
	if err != nil {
		return err
	}

	var frameNumber int

	for {
		ret = C.av_read_frame(fmtCtx, packet)
		if ret < 0 {
			break
		}
		if packet.stream_index == videoStreamIndex {
			err = packetFunc(PacketData{
				Number: frameNumber,
				Data:   slices.Clone(unsafe.Slice((*byte)(packet.data), int(packet.size))),
			})
			if err != nil {
				if errors.Is(err, io.EOF) {
					break
				}
				return err
			}

			frameNumber++
		}
		C.av_packet_unref(packet)
	}

	return nil
}
