package libopenjp2

/*
#cgo pkg-config: libopenjp2

#include "libopenjp2.h"
*/
import "C"
import (
	"encoding/binary"
	"errors"
	"fmt"
	"runtime"
	"slices"
	"unsafe"
)

//FFMpeg removed libopenjpeg decoder https://github.com/FFmpeg/FFmpeg/commit/60ccb3fe787be3bb10fc4545b3593cd1e0b769ed

const Jp2SigType = 0x6A502020
const Jp2SigValue = 0x0D0A870A

type Jpeg2000Decoder struct {
	ctx C.opj_dparameters_t
}

const QualityLayersAll = 0
const ResolutionLayersAll = 0

func NewJpeg2000Decoder(qualityLayers, resolutionLayers uint) (*Jpeg2000Decoder, error) {
	d := &Jpeg2000Decoder{}
	C.opj_set_default_decoder_parameters(&d.ctx)

	d.ctx.cp_layer = C.uint32_t(qualityLayers)
	d.ctx.cp_reduce = C.uint32_t(resolutionLayers)

	return d, nil
}

var jp2c = binary.LittleEndian.Uint32([]byte("jp2c"))

type Jpeg2000Frame struct {
	Width, Height int
	X, Y, Z       []uint32
}

func (d *Jpeg2000Decoder) DecodeFrame(buf []byte) (frame *Jpeg2000Frame, err error) {
	var image *C.opj_image_t
	var dec *C.opj_codec_t
	var stream *C.opj_stream_t

	var pinner runtime.Pinner
	pinner.Pin(unsafe.Pointer(unsafe.SliceData(buf)))
	defer pinner.Unpin()

	// Check if input is a raw jpeg2k codestream or in jp2 wrapping
	if (binary.LittleEndian.Uint32(buf) == 12) &&
		(binary.LittleEndian.Uint32(buf[4:]) == Jp2SigType) &&
		(binary.LittleEndian.Uint32(buf[8:]) == Jp2SigValue) {
		dec = C.opj_create_decompress(C.OPJ_CODEC_JP2)
	} else {
		/* If the AVPacket contains a jp2c box, then skip to
		 * the starting byte of the codestream. */
		if binary.LittleEndian.Uint32(buf[4:]) == jp2c {
			buf = buf[8:]
		}
		dec = C.opj_create_decompress(C.OPJ_CODEC_J2K)
	}

	if dec == nil {
		return nil, errors.New("error initializing decoder")
	}
	defer C.opj_destroy_codec(dec)

	// Tie decoder with decoding parameters
	C.opj_setup_decoder(dec, &d.ctx)

	stream = C.opj_stream_default_create(C.OPJ_STREAM_READ)
	defer C.opj_stream_destroy(stream)

	if stream == nil {
		return nil, errors.New("error initializing stream")
	}

	reader := &C.BufferReader{
		pos:    0,
		size:   C.int(len(buf)),
		buffer: (*C.uchar)(unsafe.Pointer(unsafe.SliceData(buf))),
	}

	pinner.Pin(reader)

	C.set_stream_callbacks(stream)
	C.opj_stream_set_user_data(stream, unsafe.Pointer(reader), nil)
	C.opj_stream_set_user_data_length(stream, C.ulong(len(buf)))

	ret := C.opj_read_header(stream, dec, &image)
	defer C.opj_image_destroy(image)

	if ret != 1 {
		return nil, errors.New("error decoding stream header")
	}

	if image.numcomps != 3 {
		return nil, fmt.Errorf("unexpected component number %d", image.numcomps)
	}

	components := unsafe.Slice(image.comps, int(image.numcomps))

	for i, c := range components {
		if c.prec != 12 {
			return nil, fmt.Errorf("unexpected component %d bit depth %d", i, c.prec)
		}
	}

	ret = C.opj_decode(dec, stream, image)

	if ret != 1 {
		return nil, errors.New("error decoding image")
	}

	for i, c := range components {
		if c.data == nil {
			return nil, fmt.Errorf("component %d has no data", i)
		}
	}

	readComponent := func(index int) []uint32 {
		return unsafe.Slice((*uint32)(unsafe.Pointer(components[index].data)), int(components[index].h)*int(components[index].w))
	}

	frame = &Jpeg2000Frame{
		Width:  int(components[0].w),
		Height: int(components[0].h),
		X:      slices.Clone(readComponent(0)),
		Y:      slices.Clone(readComponent(1)),
		Z:      slices.Clone(readComponent(2)),
	}

	return frame, nil
}
